#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


"""
Widget container as sidecar panel
"""

from ipywidgets import VBox
from traitlets import Unicode, CaselessStrEnum
from cosapp_lab._frontend import module_name, module_version


class Sidecar(VBox):
    """Widget container as sidecar panel.

    If more than one children is given, they will be appended top to bottom.

    Attributes
    ----------
    title: str, optional
        Tab title for the sidecar; default "Sidecar"
    anchor: str - one of ['split-right', 'split-left', 'split-top', 'split-bottom', 'tab-before', 'tab-after', 'right']
        Position of the sidecar; default "split-right"
    children: list of Widget
        List of widget to insert in the sidecar
    """

    _model_name = Unicode("SidecarModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("SidecarView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    title = Unicode("Sidecar").tag(sync=True)
    anchor = CaselessStrEnum(
        [
            "split-right",
            "split-left",
            "split-top",
            "split-bottom",
            "tab-before",
            "tab-after",
            "right",
        ],
        default_value="split-right",
        allow_none=False,
    ).tag(sync=True)



"""CoSApp project {{ cookiecutter.project_name }}

{{ cookiecutter.project_short_description }}
"""
from ._version import __version__


def find_resources(filename: str = "") -> str:
    """Returns the fullpath of a file in resources folder.
    
    Parameters
    ----------
    filename: str, optional
        File or directory looked for; default resources folder
    
    Returns
    -------
    str
        Full path to resources
    """
    import os
    fullpath = os.path.realpath(os.path.join(__path__[0], "resources", filename))
    if not os.path.exists(fullpath):
        raise FileNotFoundError(fullpath)
    return fullpath


from {{cookiecutter.project_slug}} import tools, ports, systems, drivers

__all__ = ["drivers", "ports", "systems", "tools"]



import logging
import uuid
import numpy

from typing import List, Optional

try:
    from OCC import Core as OCC_Core
    from OCC.Core.BRepMesh import BRepMesh_IncrementalMesh
    from OCC.Core import Poly
    from OCC.Core.TopExp import TopExp_Explorer
    from OCC.Core.BRepBndLib import brepbndlib
    from OCC.Core.Bnd import Bnd_Box
    _OCC_found = True

except ImportError as e:
    print(e)
    OCC_Core = None
    Poly = None
    BRepMesh_IncrementalMesh = None
    TopExp_Explorer = None
    _OCC_found = False

logger = logging.getLogger(__name__)


class OccParser:
    """Helper class to create a triangulation of shape."""

    def __init__(self,
        shape_list: Optional[List["OCC_Core.TopoDS.TopoDS_Shape"]] = None,
        mesh_quality = 0.75,
    ):
        if _OCC_found:
            input_shape_list = shape_list
        else:
            input_shape_list = []
            logger.warning(
                "OCC or pyoccad cannot be found; geometry viewer will not work"
            )

        self.threejs_data = {}
        self.binary_data: List[numpy.ndarray] = []
        self.binary_position = {}
        self._faces = {}
        self._faces_pos = {}
        self.shape_idx = 0

        for shape in input_shape_list:
            if isinstance(shape, OCC_Core.TopoDS.TopoDS_Shape):
                my_deviation = self.compute_default_deviation(shape)

                self.threejs_data[self.shape_idx] = []
                if self.shape_idx == 0:
                    self.binary_position[self.shape_idx] = [0, -1]
                else:
                    last = self.binary_position[self.shape_idx - 1][1]
                    self.binary_position[self.shape_idx] = [last + 1, last]
                self.threejs_data[f"edge_{self.shape_idx}"] = []
                BRepMesh_IncrementalMesh(
                    shape, my_deviation * mesh_quality, False, 0.5 * mesh_quality, True
                )
                self.__build_face_mesh(shape)
                self.__build_edge_mesh(shape)
                self.shape_idx += 1

            elif isinstance(shape, dict):
                self.threejs_data[self.shape_idx] = []
                if self.shape_idx == 0:
                    self.binary_position[self.shape_idx] = [0, -1]
                else:
                    last = self.binary_position[self.shape_idx - 1][1]
                    self.binary_position[self.shape_idx] = [last + 1, last]
                self.threejs_data[f"edge_{self.shape_idx}"] = []
                self.threejs_data[f"misc_{self.shape_idx}"] = shape.get("misc_data", {})
                if shape["shape"] is not None:
                    BRepMesh_IncrementalMesh(shape["shape"], 0.005, True, 0.5, True)
                    color = shape.get("color", "#156289")
                    trans = shape.get("transparent", False)
                    self.__build_face_mesh(shape["shape"], color, trans)
                    render_edge = shape.get("edge", False)
                    if render_edge:
                        self.__build_edge_mesh(shape["shape"])

                self.shape_idx += 1

    def __build_face_mesh(
        self,
        sh: "OCC_Core.TopoDS.TopoDS_Shape",
        color: str = "#156289",
        transparent: bool = False,
    ):
        """
        Create a triangle mesh of surface of input shape. This is a simplified
        version of the corresponding function in Pyoccad

        Parameters
        ----------
        sh : OCC_Core.TopoDS.TopoDS_Shape
            Input shape to me meshed.

        """
        expl = TopExp_Explorer(sh, OCC_Core.TopAbs.TopAbs_FACE)
        while expl.More():
            face = OCC_Core.TopoDS.topods.Face(expl.Current())
            self.__addFaceMesh(face, color, transparent)
            expl.Next()

    def __addFaceMesh(self,
        face: "OCC_Core.TopoDS.TopoDS_Face",
        color: str,
        transparent: bool,
    ):
        """
        Helper function to create a triangle mesh of input surface . This is a simplified
        version of the corresponding function in Pyoccad

        Parameters
        ----------
        face : OCC_Core.TopoDS.TopoDS_Face
            Input shape to me meshed.
        """
        loc = OCC_Core.TopLoc.TopLoc_Location()
        T = OCC_Core.BRep.BRep_Tool().Triangulation(face, loc)
        if T is None:
            return

        vtx = []
        for node in self.get_nodes(T):
            vtx.extend(node.Coord(d) for d in range(1, 4))
        idx = []
        for triangle in self.get_triangles(T):
            idx.extend(triangle.Value(d) - 1 for d in range(1, 4))

        faces = numpy.array(idx, "uint16")
        trf = loc.Transformation()
        translation = trf.TranslationPart()
        quaternion = trf.GetRotation()
        self.threejs_data[self.shape_idx].append(
            {
                # "vertices": vtx,
                # "faces": faces.tolist(),
                "pos": [
                    translation.X(),
                    translation.Y(),
                    translation.Z(),
                ],
                "quat": [
                    quaternion.X(),
                    quaternion.Y(),
                    quaternion.Z(),
                    quaternion.W(),
                ],
                "color": color,
                "transparent": transparent,
            }
        )
        self.binary_data.append(numpy.array(vtx, dtype="float64"))
        self.binary_data.append(faces)
        self.binary_position[self.shape_idx][1] += 2
        mesh_id = uuid.uuid4().hex
        self._faces[mesh_id] = face
        self._faces_pos[mesh_id] = vtx

    def __build_edge_mesh(self, sh: "OCC_Core.TopoDS.TopoDS_Shape"):
        """
        Create a the edge mesh input shape. This is a simplified
        version of the corresponding function in Pyoccad

        Parameters
        ----------
        sh : OCC_Core.TopoDS.TopoDS_Shape
            Input shape to me meshed.
        """
        edgeMap = OCC_Core.TopTools.TopTools_IndexedDataMapOfShapeListOfShape()
        OCC_Core.TopExp.topexp.MapShapesAndAncestors(
            sh, OCC_Core.TopAbs.TopAbs_EDGE, OCC_Core.TopAbs.TopAbs_FACE, edgeMap
        )

        for i in range(1, edgeMap.Size() + 1):
            faceList = edgeMap.FindFromIndex(i)
            if faceList.Size() != 0:
                face = OCC_Core.TopoDS.topods.Face(faceList.First())
                edge = OCC_Core.TopoDS.topods.Edge(edgeMap.FindKey(i))
                # Looking for face mesh to recover position buffer and save memory
                for mesh_id, f in self._faces.items():
                    if face.IsSame(f):
                        vertexBuffer = self._faces_pos.get(mesh_id, None)

                if vertexBuffer is not None:
                    loc = OCC_Core.TopLoc.TopLoc_Location()
                    tri = OCC_Core.BRep.BRep_Tool().Triangulation(face, loc)
                    trf = loc.Transformation()
                    translation = trf.TranslationPart()
                    quaternion = trf.GetRotation()
                    # polygon = Poly.PolygonOnTriangulation(edge, face, loc)
                    polygon = OCC_Core.BRep.BRep_Tool().PolygonOnTriangulation(edge, tri, loc)

                    self.threejs_data[f"edge_{self.shape_idx}"].append(
                        {
                            "vertices": vertexBuffer,
                            "faces": list(range(polygon.NbNodes())),
                            "pos": [
                                translation.X(),
                                translation.Y(),
                                translation.Z(),
                            ],
                            "quat": [
                                quaternion.X(),
                                quaternion.Y(),
                                quaternion.Z(),
                                quaternion.W(),
                            ],
                        }
                    )

    def compute_default_deviation(self, shape) -> float:
        """Compute the minimum size of `shape` bounding box.
        This function is used to compute the mesh quality in tesselation.
        """
        box = Bnd_Box()
        brepbndlib.Add(shape, box)
        x_min, y_min, z_min, x_max, y_max, z_max = box.Get()
        return max(x_max - x_min, y_max - y_min, z_max - z_min) * 2e-2

    @staticmethod
    def get_nodes(T):
        for i in range(T.NbNodes()):
            yield T.Node(i + 1)

    @staticmethod
    def get_triangles(T):
        yield from T.Triangles()



import copy
import json
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from weakref import ref
import numbers
import numpy
from cosapp.drivers import Driver
from cosapp.drivers.time.interfaces import ExplicitTimeDriver
from cosapp.ports.enum import PortType
from cosapp.ports.port import Variable, BasePort
from cosapp.systems import System
from cosapp.drivers import NonLinearSolver
from .cosapp_parser import CosappParser
from .utils import is_jsonable, replicate_dict_structure
from numpy import linalg


class CosappObjectParser(CosappParser):
    """Class to read/modifier/interact with  cosapp system

    Parameters
    ----------
    system : cosapp.systems.System
    """

    def __init__(self, system: System) -> None:

        super().__init__(system)

    def set_up_system(self, data: System) -> None:
        self.system_name = data.name
        self._system = ref(data)
        try:
            self.system_dict = replicate_dict_structure(data.to_dict())
        except:
            self.system_dict = {}

        self._discover_children(data)
        self._get_data_from_system()
        for sys_name in self.children_list:
            self._discover_driver(sys_name)

    def get_system_from_name(self, full_sys_path: str) -> System:
        """Return an instance of system from system name

        Parameters
        ----------
        sys_name : str
            Name of requested system

        Returns
        -------
        system : cosapp.systems.System

        """
        # Since full_sys_path starts with the name of system but
        # the system path used in __getattr__ method of CosApp system
        # does not take into account the name of system, we need to
        # check if full_sys_path is system name or not.
        if full_sys_path == self.system_name:
            current_system = self._system()
        else:
            sys_path = ".".join(full_sys_path.split(".")[1:])
            current_system = self._system()[sys_path]
        return current_system

    def get_children_var(self, sys_name: str, port_name: str) -> List[str]:
        """Get all variables in a port of a system.

        Parameters
        ----------
        sys_name :str
            Name of system

        port_name : str
            Name of port

        Returns
        -------
        List[str]
            List of variables inside the requested port
        """
        try:
            ret = self._children[sys_name]["port_data"][port_name]
        except:
            ret = []
        return ret

    def get_children_var_input(
        self, sys_name: str, port_name: str, check_input: bool = False
    ) -> Dict[str, Dict]:
        """Get the variables of a port which is not a sink of any
        connection

        Parameters
        ----------
        sys_name :str
            Name of system

        port_name : str
            Name of port

        Returns
        -------
        Dict[str, Dict]
            A dictionary with keys are the variables of system which are not a sink of any
        connection and values are the JSON data of these variables.
        """
        current_system = self.get_system_from_name(sys_name)
        current_port = getattr(current_system, port_name)
        try:
            var_list = self._children[sys_name]["port_data"][port_name]
            if check_input:
                input_list = self._children[sys_name]["port_data_input"].get(
                    port_name, []
                )
            else:
                input_list = []
            ret = [x for x in var_list if x not in input_list]
        except:
            ret = []

        if len(ret) == 0:
            return {"Mutable variable not found": {}}
        else:
            var_dict = {}
            for var_name in ret:
                var = current_port._variables[var_name]
                var_dict[var_name] = var.__json__()
            return var_dict

    def serialize_data_from_system(self, dumps=True) -> Union[str, Dict]:
        """Serialize all values of variables in current system if possible.

        Parameters
        ----------
        dumps : boolean
            Flag to check if the return object is a Dict or a Json string

        Returns
        -------
        Union[str, Dict[str, Any]]
            A dictionary with keys are the contextual name of the variables, value
            are a list of 2 value, first one is de type of data and second one is
            the value of data.
        """
        content = {}
        system = self._system()
        for var_name in self.variable_list:
            value = system[var_name]
            if is_jsonable(value):
                pkl_str = value
            elif isinstance(value, numpy.ndarray):
                pkl_str = value.tolist()
            else:
                pkl_str = "non-jsonable"
            key = f"{system.name}.{var_name}"
            typename = type(value).__name__
            content[key] = [typename, pkl_str]
        
        return json.dumps(content) if dumps else content

    def serialize_recorder(self) -> Dict:
        """Serialize all dataframe recorder in system.

        Returns
        -------
        Dict
            A dictionary with keys are the path to the recorder, in form
            of `system.driver.recorder` and value is the Json data of the
            dataframe.
        """

        ret = {}
        for sys_name in self._children:
            current_system = self.get_system_from_name(sys_name)
            for driver_data in self._driver[sys_name].values():
                driver_path = driver_data["path"]
                full_path = ".".join(driver_path)
                current_driver = current_system.drivers[driver_path[0]]
                for element in driver_path[1:]:
                    current_driver = current_driver.children[element]
                if (
                    current_driver.recorder is not None
                    and current_driver.recorder.includes != ["_"]
                ):
                    try:
                        recorder_frame = current_driver.recorder.export_data()
                    except:
                        recorder_frame = current_driver.recorder.data
                    json_data = json.loads(recorder_frame.to_json())
                    for key, value in json_data.items():
                        temp = []
                        for temp_val in value.values():
                            if isinstance(temp_val, numbers.Number):
                                temp.append([temp_val])
                            else:
                                temp.append(temp_val)
                        json_data[key] = temp

                    ret[f"{sys_name}.{full_path}"] = json_data

        return ret

    def serialize_driver_data(self) -> Dict:
        """Serialize all data related to a NonLinerSolver. In order
        to catch the residue vector, the `history` flag of solver need
        to be `True`

        Returns
        -------
        Dict
            A dictionary with keys are the path to the driver, in form
            of `system.driver` and value is the residue vector of the solver.
        """

        ret = {}
        for sys_name in self._children:
            current_system = self.get_system_from_name(sys_name)
            for driver_data in self._driver[sys_name].values():
                driver_path = driver_data["path"]
                full_path = ".".join(driver_path)
                current_driver = current_system.drivers[driver_path[0]]
                for element in driver_path[1:]:
                    current_driver = current_driver.children[element]
                if isinstance(current_driver, NonLinearSolver):
                    trace = current_driver._NonLinearSolver__trace
                    if len(trace) > 0:
                        ret[f"{sys_name}.{full_path}"] = {}
                        residue_vec = []
                        for item in trace:
                            residue_vec.append(linalg.norm(item["residues"]))
                        ret[f"{sys_name}.{full_path}"]["Residue"] = residue_vec

        return ret

    def _get_data_from_system(self, serializable=False) -> Dict:
        """Read all values of variables in current system
        CosappObjectParser saves the value of all variables in order to
        restore the system to its initial state.
        """
        self.system_variable_data = {}
        for sub_system in self._children:
            sys_root = (
                ""
                if sub_system == self.system_name
                else ".".join(self._children[sub_system]["path"].split(".")[1:])
            )
            port_list = self._children[sub_system]["port_list"]
            for port in port_list:
                for variable_name in self._children[sub_system]["port_data"][port]:
                    full_var_path = "{}.{}.{}".format(sys_root, port, variable_name)
                    self.variable_list.append(full_var_path.strip("."))
                    if port != System.INWARDS and port != System.OUTWARDS:
                        var_path = full_var_path
                    else:
                        var_path = "{}.{}".format(sys_root, variable_name)
                    var_path = var_path.strip(".")
                    try:
                        value = copy.deepcopy(self._system()[var_path])
                        try:
                            size = len(value)
                            if serializable:
                                if isinstance(value, numpy.ndarray):
                                    value = self._system()[var_path].tolist()
                        except:
                            size = 1
                    except Exception as e:
                        value = None
                        size = 1
                    self.system_variable_data[var_path] = {"size": size, "value": value}
        return self.system_variable_data

    def reset_variable_value(self) -> None:
        """Reset system to its initial state"""
        for path in self.system_variable_data:
            value = self.system_variable_data[path]["value"]
            if value is not None:
                self._system()[path] = copy.deepcopy(value)

    def _discover_driver(
        self,
        sys_name: str,
        current_drive: Driver = None,
        parent_driver: List[str] = None,
    ) -> None:
        """Get the all drivers of system and its sub-systems recursively

        Parameters
        ----------
        sys_name : str
            The name of system in question

        current_driver : cosapp.drivers.Driver
            The driver of current system

        parent_driver : List[str]
            List of the parents driver of current driver

        Returns
        -------
        None

        """

        current_system = self.get_system_from_name(sys_name)

        if sys_name not in self._driver:
            self._driver[sys_name] = {}
            content = {}
        else:
            content = self._driver[sys_name]
        if current_drive is None:
            driver_dict = current_system.drivers
            chidren_driver_list = list(current_system.drivers)
            base_path_list = []
        else:

            driver_dict = current_drive.children
            chidren_driver_list = list(current_drive.children)
            base_path_list = [x for x in parent_driver]

        for children_driver in chidren_driver_list:
            path_list = base_path_list + [children_driver]
            content[children_driver] = {"path": path_list}
            if driver_dict[children_driver].recorder is not None:
                content[children_driver]["recorder"] = path_list
                content[children_driver]["variables"] = driver_dict[
                    children_driver
                ].recorder.field_names()
            else:
                if current_drive is not None and current_drive.recorder is not None:
                    parent_rec_var = current_drive.recorder.field_names()
                    content[children_driver]["recorder"] = (
                        base_path_list if len(parent_rec_var) > 0 else False
                    )
                    content[children_driver]["variables"] = (
                        parent_rec_var if len(parent_rec_var) > 0 else []
                    )
                else:
                    content[children_driver]["recorder"] = False
                    content[children_driver]["variables"] = []

        self._driver[sys_name] = content

        for children_driver in chidren_driver_list:
            path_list = base_path_list + [children_driver]
            self._discover_driver(sys_name, driver_dict[children_driver], path_list)

    def connect_main_driver(self, f: Callable) -> None:
        """Connect the computed signal of first driver of system to slot f

        Parameters
        ----------
        f :Callable
            The slot to be connected
        """
        key_list = list(self._system().drivers)

        if len(key_list) > 0:
            self._system().drivers[key_list[0]].computed.connect(f)
        else:
            self._system().computed.connect(f)

    def run_system(self) -> None:
        """Helper function to run driver"""

        self._system().run_drivers()

    def check_validity(self) -> Dict:
        """ Call the validation check of input system"""
        return self._system().check()

    def set_variable_value(
        self, sys_name: str, port: str, variable: str, value: Any
    ) -> None:
        """Modify the value of variable inside a port of system

        Parameters
        ----------
        sys_name :str
            Name of system

        port : str
            Name of port

        variable : str
            Name of variable

        value : Any
            Value of variable to be modified
        """
        current_system = self.get_system_from_name(sys_name)
        current_system[port][variable] = value

    def _discover_children(self, system: System, parent: Optional[str] = None) -> None:
        """Get the sub system of input system

        Parameters
        ----------
        system : cosapp.systems.System
            The input system

        parent: Optional[str]
            Path in form of "parent.child.sub-child ..." to the parent of input system.
            parent = None if input system does not have parent.

        Returns
        -------
        Dict[str, List[str]]
            Dictionary of sub system of input system with their children.
        """
        if parent is not None:
            key = parent + "." + system.name
        else:
            key = system.name

        if key in self._children:
            content = self._children[key]
        else:
            content = {}
        if len(system.children) == 0:
            content["children"] = None
        else:
            content["children"] = list(system.children)

        child_name = ".".join(key.split(".")[1:])
        content["port_data"] = {}
        content["port_list"] = []
        content["in_port_list"] = []
        content["out_port_list"] = []
        content["path"] = key
        if child_name != "":
            current_system = self._system()[child_name]
        else:
            current_system = self._system()

        ports = list(current_system.inputs.values()) + list(
            current_system.outputs.values()
        )
        for port in ports:
            if port.direction == PortType.IN:
                content["in_port_list"].append(port.name)
            elif port.direction == PortType.OUT:
                content["out_port_list"].append(port.name)

            content["port_list"].append(port.name)
            content["port_data"][port.name] = list(port.serialize_data())

        content["driver_list"] = list(current_system.drivers)

        if key not in self._children:
            self._children[key] = content
        else:
            for content_key in content:
                self._children[key][content_key] = content[content_key]
        for c_name in system.children:
            self._discover_children(system[c_name], key)

    def get_time_driver(self) -> List[ExplicitTimeDriver]:
        """Return the list of ExplicitTimeDriver in system"""
        ret = []
        for sys_name in self.children_list:
            for driver_name in self.children_drive.get(sys_name, []):
                if driver_name != "None":
                    driver_path = self._driver[sys_name][driver_name]["path"]
                    current_system = self.get_system_from_name(sys_name)
                    for index, path in enumerate(driver_path):
                        if index == 0:
                            selected_driver = current_system.drivers[path]
                        else:
                            selected_driver = selected_driver.children[path]
                    if isinstance(selected_driver, ExplicitTimeDriver):
                        ret.append(selected_driver)
        return ret

    def get_recorder_variable(
        self, sys_name: str, driver_name: str
    ) -> Tuple[List[str], List[int]]:
        """Get the list of variable and size of each in the recorder of driver

        Parameters
        ----------
        sys_name :str
            Name of system

        driver_name : str
            Name of driver

        Returns
        -------
        Tuple[List[str], List[int]]
            The first element of the tuple contains the list of variable name
            inside the selected recorder. The second one contains the size of
            each variable, which is 1 for scalar variable and length of vector for
            vector valued variable.

        """
        if len(self._driver[sys_name]) == 0:
            return ["None"], [1]
        elif self._driver[sys_name][driver_name]["recorder"]:
            var_size = []
            ret = self._driver[sys_name][driver_name]["variables"]
            for var_name in ret:
                var_size.append(self.system_variable_data[var_name]["size"])
            return ret, var_size
        else:
            ret = ["None"], [1]
            return ret

    def get_data_from_recorder(
        self, sys_name: str, driver_name: str, var_name: str
    ) -> Tuple[str, Union[str, List[float]], List[float]]:
        """Get value of variable in recorder, depend on the type of driver
        (ExplicitTimeDriver or not), the content of return with be different.
        The tuple returned has the following structure:
        Tuple[str, List, List ]:
            If driver is ExplicitTimeDriver:
                ("time", [Time reference], [Value of variable])
            else :
                ("static", Driver name, [Value of variable]])

        Parameters
        ----------
        sys_name :str
            Name of system

        driver_name : str
            Name of driver

        var_name : str
            Name of variable

        Returns
        -------
        Tuple[str, Union[str, List[float]], List[float]]

        """

        current_system = self.get_system_from_name(sys_name)

        recorder_path = self._driver[sys_name][driver_name]["recorder"]
        current_driver = current_system.drivers[recorder_path[0]]
        for element in recorder_path[1:]:
            current_driver = current_driver.children[element]
        try:
            recorder_data = (
                current_driver.recorder.export_data()
            )  # use new API to get data from recorder
        except:
            recorder_data = current_driver.recorder.data
        if isinstance(current_driver, ExplicitTimeDriver):
            reference = numpy.array(recorder_data["Reference"], dtype=float)
            try:
                value = numpy.array(recorder_data[var_name], dtype=float)
            except ValueError:
                value = recorder_data[var_name].values
            ret = ("time", reference.tolist(), value.tolist())

        else:
            reference = list(recorder_data["Reference"])
            if driver_name in reference:
                index = reference.index(driver_name)
                value = recorder_data[var_name][index]
                if isinstance(value, list):
                    ret = ("static", driver_name, value)
                else:
                    ret = ("static", driver_name, [value])

        return ret

    @property
    def children_port_meta(self) -> Dict[str, Dict]:
        """Get a dictionary which contains system name
        as key and the metadata of port as value

        Returns
        -------
        Dict[str, Dict[str, Any]]
            Dict contains name of port of each child in system, including itself
        """
        ret = {}
        for sys, data in self._children.items():
            system = self.get_system_from_name(sys)
            ret[sys] = {}
            for port_name in data["port_list"]:
                port: BasePort = system[port_name]
                ret[sys][port_name] = self.port_to_dict(port)

        return ret

    def port_to_dict(self, port: BasePort) -> Dict:
        """
        Convert a CoSApp port to a python dict
        """
        tmp = dict()
        if port.name not in ["inwards", "outwards"]:
            tmp["__class__"] = port.__class__.__qualname__
            for variable_name, variable in port._variables.items():
                tmp[variable_name] = self.variable_to_dict(variable)
        else:
            for v_name, variable in port._variables.items():
                tmp[v_name] = self.variable_to_dict(variable)
        return tmp

    def variable_to_dict(self, var: Variable) -> Dict:
        """Convert input variable into a dictionary.

        Returns
        -------
        dict
            The dictionary representing this variable.
        """
        def get_attr(name, transform = lambda a: a or None):
            try:
                attr = getattr(var, name)
            except AttributeError:
                return None
            return transform(attr)

        result = {}
        mapping = {
            "unit": "unit",
            "desc": "description",
            "invalid_comment": "invalid_comment",
            "out_of_limits_comment": "out_of_limits_comment",
        }
        for key, name in mapping.items():
            result[key] = get_attr(name)
        
        result["distribution"] = get_attr(
            "distribution",
            lambda d: d.__json__() if d else None
        )

        for key in ["valid_range", "limits"]:
            try:
                range_attr = list(getattr(var, key))
            except:
                range_attr = None
            else:
                range_attr = list(map(
                    lambda x: x if numpy.isfinite(x) else None,
                    range_attr
                ))
                if range_attr == [None, None]:
                    range_attr = None
            result[key] = range_attr

        # Filter out None values
        # return dict(filter(lambda it: it[1] is not None, result.items()))
        return result



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.

from typing import Any, Callable, Dict, List
from weakref import ReferenceType
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser
from cosapp_lab.widgets.base.base_component import BaseComponent



class StructureComponent(BaseComponent):
    name = "StructureView"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List):
        if content["action"] == f"{self.name}::getData":
            try:
                from cosapp.tools import VisJsRenderer

            except ImportError:
                self.send(
                    {
                        "type": f"{self.name}::importError",
                    }
                )
            else:
                visJsData = VisJsRenderer(self.system)
                self.send(
                    {
                        "type": f"{self.name}::structureData",
                        "payload": visJsData.get_data(),
                    }
                )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from cosapp.systems import System
from weakref import ReferenceType
from typing import Callable, Dict, List, Any
from cosapp_lab.widgets.utils import CosappObjectParser


class BaseComponent:

    name = "BaseComponent"

    def __init__(
        self,
        data: "ReferenceType[System]" = None,
        sys_data: CosappObjectParser = None,
        send_func: Callable = None,
        **kwargs
    ):

        self._system = data
        self.sys_data = sys_data
        self.send = send_func

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List) -> None:
        pass

    def computed_notification(self) -> None:
        pass

    @property
    def system(self) -> System:
        return self._system()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


"""
Widget container as SysExp panel
"""
import copy
import json
import logging
import os
import re

from collections import OrderedDict
from datetime import datetime
from io import StringIO
from typing import Dict, List, NoReturn, Union, Any
from weakref import ReferenceType, ref

import ipykernel.connect as kernel
import numpy as np
import requests
from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp_lab._frontend import module_name, module_version
from cosapp_lab.widgets.utils import (
    CosappJsonParser,
    CosappObjectParser,
    OccParser,
    get_nonexistant_path,
    is_jsonable,
)
from ipywidgets import Box
from traitlets import CaselessStrEnum
from traitlets import Dict as tDict
from traitlets import Int, Unicode, observe

COSAPP_URL_START = "cosapp/server/start"
COSAPP_URL_STOP = "cosapp/server/stop"
COSAPP_URL_RUN = "cosapp/server/run"


class SysExplorer(Box):
    """Widget container as SysExplorer panel.

    If more than one children is given, they will be appended top to bottom.

    Attributes
    ----------
    title: str, optional
        Tab title for the SysExplorer; default "SysExplorer"
    anchor: str - one of ['split-right', 'split-left', 'split-top', 'split-bottom', 'tab-before', 'tab-after', 'right']
        Position of the SysExplorer; default "split-right"
    children: list of Widget
        List of widget to insert in the SysExplorer
    """

    _model_name = Unicode("SysExplorerModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("SysExplorerView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    title = Unicode("SysExplorer").tag(sync=True)
    anchor = CaselessStrEnum(
        [
            "split-right",
            "split-left",
            "split-top",
            "split-bottom",
            "tab-before",
            "tab-after",
            "right",
        ],
        default_value="tab-after",
        allow_none=False,
    ).tag(sync=True)
    system_data = tDict(default_value={"key": "None"}, allow_none=False).tag(sync=True)
    system_dict = tDict(default_value={"key": "None"}, allow_none=False).tag(sync=True)
    geo_data = tDict(default_value={"key": "None"}, allow_none=False).tag(sync=True)
    computed_data = tDict(default_value={"key": "None"}, allow_none=False).tag(
        sync=True
    )
    recorder_data = tDict(default_value={"key": "None"}, allow_none=False).tag(
        sync=True
    )
    driver_data = tDict(default_value={"key": "None"}, allow_none=False).tag(sync=True)
    progress_geo_update = tDict(default_value={"key": "None"}, allow_none=False).tag(
        sync=True
    )
    update_signal = Int(default_value=0, allow_none=False).tag(sync=True)
    notification_msg = tDict(
        default_value={"update": 0, "msg": "", "log": ""}, allow_none=False
    ).tag(sync=True)
    initial_store = tDict(default_value={}, allow_none=False).tag(sync=True)

    server_msg = tDict(
        default_value={"update": 0, "msg": "", "log": ""}, allow_none=False
    ).tag(sync=True)

    system_config = tDict(default_value={}, allow_none=False).tag(sync=True)

    def __init__(self, data: Union[System, Dict] = None, **kwargs):
        """Initialize class from input system

        Attributes
        ----------
        self._system : ReferenceType
            The weak reference of input system

        self.add_shape : Callable
            The helper function used to get geometry from system.
            If it is not provided, self.default_add_shape will
            be used.

        self.title : str
            The title of widget tab.

        self.update_signal : int
            The signal which is observed in frontend, it is incremented each time
            the main driver is computed

        self.time_step : int
            The counter for time step

        self.geo_data : Dict
            The dictionary which holds the parsed geometry data.

        self.notification_msg : Dict
            The message which is sent to frontend after each time step.

        """

        if isinstance(data, System):
            self._system: ReferenceType[System] = ref(data)
            data_name = data.name
            self.system_config = {"mode": "run"}
        elif isinstance(data, Dict):
            self._system: ReferenceType[System] = ref(System("temp"))
            data_name = list(data["Systems"].keys())[0]
            self.system_config = {"mode": "edit"}
        else:
            return
        self.geo_source = kwargs.get("source", None)
        self.add_shape = kwargs.get("add_shape", None)
        self.initial_store = kwargs.get("state", {"systemConfig": self.system_config})
        # self.title = f"{data_name} - SysExplorer {module_version}"
        self.update_signal = 0
        self.time_step = 0
        self.geo_data = OrderedDict()
        self.computed_data = {}
        self.recorder_data = {}
        self.driver_data = {}
        self.progress_geo_update = {}
        self.local_buffer = OrderedDict()
        self.server_process = None
        self.server_log = ""
        self.__init_data(data)
        self.__init_connection()

        self.notification_msg = {"update": 0, "msg": "", "log": ""}
        self.cosapp_thread = None
        super().__init__(**kwargs)

    @observe("system_data")
    def handle_data_change(self, change):
        """
        Not implemented yet
        """
        pass

    def __init_data(self, data: System):
        """Parse cosapp System and return list of sub-systems,
        list of ports...

        Parameters
        ----------
        data : cosapp.systems.System
            Input system
        """
        self.system_variable = {}
        if isinstance(data, System):
            self.sys_data = CosappObjectParser(data)
        elif isinstance(data, Dict):
            self.sys_data = CosappJsonParser(data)

        self.system_dict = self.sys_data.flattened_system
        # self._system = ref(data)
        self._system_list = self.sys_data.children_list
        self._driver_list = self.sys_data.children_drive
        in_port_dict = self.sys_data.children_in_port
        out_port_dict = self.sys_data.children_out_port
        self.systemGraphData = {}
        for key in self._system_list:
            try:
                connection_list = self.system_dict[key]["connections"]
            except:
                connection_list = []
            self.systemGraphData[key] = {
                "inPort": in_port_dict[key],
                "outPort": out_port_dict[key],
                "connections": connection_list,
            }

        port_dict = self.sys_data.children_port
        for sys_name in port_dict:
            for port_name in port_dict[sys_name]:
                variable_dict = self.sys_data.get_children_var_input(
                    sys_name, port_name
                )
                if "Mutable variable not found" not in variable_dict:
                    for var_name in variable_dict:
                        if isinstance(variable_dict[var_name]["value"], np.ndarray):
                            variable_dict[var_name]["value"] = variable_dict[var_name][
                                "value"
                            ].tolist()
                            self.system_variable[
                                f"{sys_name}.{port_name}.{var_name}"
                            ] = variable_dict[var_name]
                        elif is_jsonable(variable_dict[var_name]["value"]):
                            self.system_variable[
                                f"{sys_name}.{port_name}.{var_name}"
                            ] = variable_dict[var_name]

        if self.system_config["mode"] == "run":
            computedResult = self.sys_data.serialize_data_from_system(False)
            recorderData = self.sys_data.serialize_recorder()
            driverData = self.sys_data.serialize_driver_data()
        else:
            recorderData = {}
            computedResult = {}
            driverData = {}
        self.system_data = {
            "systemGraph": {
                "systemGraphData": self.systemGraphData,
                "systemList": self._system_list,
                "graphJsonData": {},
            },
            "systemPBS": {},
            "systemTree": self.sys_data.tree_dict,
            "variableData": self.system_variable,
            "computedResult": computedResult,
            "recorderData": recorderData,
            "driverData": driverData,
        }
        geo_data = self.get_geometry(self._system())
        threejs_data = geo_data.threejs_data
        if len(threejs_data) > 0:
            self.geo_data[0] = threejs_data

    def __init_connection(self) -> None:
        """Initialize the connection between fontend - backend and
        between input system with the callbacks.
        """
        self.sys_data.connect_main_driver(self.computed_notification)
        self.on_msg(self._handle_button_msg)
        self.old_log = ""
        self.log_stream = StringIO()
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(stream=self.log_stream)
        handler.setFormatter(logging.Formatter(fmt="%(levelname)s:%(message)s"))
        logger.addHandler(handler)

        if self.geo_source is not None and "recorder" in self.geo_source:
            self._static = False
            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                recorder = self._system().drivers[driver_name].recorder
            else:
                recorder = (
                    self._system()[".".join(path_list[0:-1])]
                    .drivers[driver_name]
                    .recorder
                )
            recorder.state_recorded.connect(self.save_data)
        else:
            driver_list = self.sys_data.get_time_driver()
            if len(driver_list) > 0:
                self._static = False
                time_driver = driver_list[0]
                if time_driver.recorder is not None:
                    time_driver.recorder.state_recorded.connect(self.save_data)
                else:
                    notification_recorder = time_driver.add_recorder(
                        DataFrameRecorder(includes=["_"])
                    )
                    notification_recorder.state_recorded.connect(self.save_data)
            else:
                self._static = True

    def computed_notification(self) -> None:
        """Callback function used to update geometry data from buffer
        and to emit update signal to front end.

        This function is called after the main driver finished. The
        geometry data is transferred from local buffer into `self.geo_data`,
        this traitlet is synchronized with the frontend.
        The traitlet `self.update_signal` is incremented to notify the fontend
        about the termination signal.
        """

        if self._static:
            self.save_data(time_ref=0)
        self.geo_data = self.local_buffer
        self.computed_data = self.sys_data.serialize_data_from_system(False)
        self.recorder_data = self.sys_data.serialize_recorder()
        self.driver_data = self.sys_data.serialize_driver_data()
        self.update_signal = self.update_signal + 1
        self.server_log = self.log_stream.getvalue()
        self.log_stream.truncate(0)
        self.log_stream.seek(0)
        self.local_buffer = OrderedDict()
        self.time_step = 0

    def update_server_log(self, msg: str) -> NoReturn:
        """Update log message of server in frontend"""
        current_update = self.server_msg["update"]
        now = datetime.now().strftime("[%H:%M:%S]")
        self.server_msg = {
            "update": current_update + 1,
            "msg": "update",
            "log": f"{now} {msg} \n",
        }

    def start_server(self, token: str, base_url: str):
        """Create and then start a server process for `cosapp_server`.
        Current cosapp system is transferred to the server through
        `server_data.json` file.

        """
        if self.server_process is None and COSAPP_URL_START is not None:
            connection_file_path = kernel.get_connection_file()
            system_name = self._system().name
            data = {
                "token": token,
                "connection": connection_file_path,
                "system_name": system_name,
            }
            start_url = base_url + COSAPP_URL_START
            try:
                proxies = {"http": None, "https": None}

                r = requests.post(start_url, json=data, proxies=proxies)
                if r.text == "1":
                    self.server_process = {"token": token, "base_url": base_url}
                    server_update = self.server_msg["update"]
                    self.server_msg = {
                        "update": server_update + 1,
                        "msg": "ok",
                        "log": "",
                    }
            except Exception as e:
                server_update = self.server_msg["update"]
                self.server_msg = {
                    "update": server_update + 1,
                    "msg": "error",
                    "log": str(e),
                }

    def stop_server(self):
        """Stop running cosapp server and remove `server_data.json` file."""
        if self.server_process is not None:
            data = {"token": self.server_process["token"]}
            stop_url = self.server_process["base_url"] + COSAPP_URL_STOP
            try:
                proxies = {"http": None, "https": None}
                r = requests.post(stop_url, json=data, proxies=proxies)
                if r.text == "1":
                    self.server_process = None
            except Exception as e:
                print(e)

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List) -> None:
        """Helper function to receive message from front end.

        Parameters
        ----------
        model : SysExplorerModel
            Current model value.

        content : {"action" : "runSignal" | "requestUpdate", "payload" : Dict}
            If action == "runSignal" => run the main
            driver of system, payload contains a dict of system parameters
            and its value.
            If action == "requestUpdate" => Send the geometry of
            the last computed step to front-end to update the computing
            progress.

        buffers : List

        """
        if content["action"] == "runSignal":
            self.sys_data.reset_variable_value()
            for key, variable_value in content["payload"].items():
                idx_group = re.search(r"\[(.*?)\]", key)
                if idx_group is not None:
                    var_idx = int(idx_group.group(1))
                    var_key = key.replace(idx_group.group(0), "")
                    sys_path = ".".join(var_key.split(".")[:-2])
                    port = var_key.split(".")[-2]
                    variable_name = var_key.split(".")[-1]

                    selected_system = self.sys_data.get_system_from_name(sys_path)
                    var_value = copy.deepcopy(selected_system[port][variable_name])
                    var_value[var_idx] = variable_value
                    self.sys_data.set_variable_value(
                        sys_path, port, variable_name, var_value
                    )
                else:
                    sys_path = ".".join(key.split(".")[:-2])
                    port = key.split(".")[-2]
                    variable_name = key.split(".")[-1]

                    self.sys_data.set_variable_value(
                        sys_path, port, variable_name, variable_value
                    )

            self.time_step = 0
            # clear() won't work here because we need a new reference for  self.geo_data
            # in order to notify the font end about the change of data.
            self.geo_data = OrderedDict()
            self.local_buffer = OrderedDict()
            self.server_log = ""

            if "currentThread" in content:
                self.sys_data.run_system()
            else:
                # if self.cosapp_thread is not None and self.cosapp_thread.is_alive():
                #     self.cosapp_thread.join()
                # self.cosapp_thread = threading.Thread(target=self.sys_data.run_system)
                # self.cosapp_thread.start()
                self.sys_data.run_system()

        elif content["action"] == "requestUpdate":
            key_list = list(self.local_buffer)
            if len(key_list) > 0:
                self.progress_geo_update = {
                    "data": self.local_buffer[key_list[-1]],
                    "time_step": key_list[-1],
                }

        elif content["action"] == "switchServer":
            request_token = content["payload"]["token"]
            request_url = content["payload"]["url"]
            if content["payload"]["signal"]:
                self.start_server(request_token, request_url)
            else:
                self.stop_server()
        elif content["action"] == "requestComputedNotification":
            self.computed_notification()
        elif content["action"] == "chartViewerSaveJson":
            json_name = content["payload"]["jsonName"].replace(".json", "")
            file_path = get_nonexistant_path(
                os.path.join(os.getcwd(), f"{json_name}.json")
            )
            json_data = content["payload"]["jsonData"]
            with open(file_path, "w") as f:
                json.dump(json_data, f)

        elif content["action"] == "requestInitialGeometry":
            self.send_initial_geometry()

    def send_initial_geometry(self) -> None:
        pass

    @staticmethod
    def default_add_shape(r: List, system: System) -> None:
        """Helper function used to extract all geometry inside
        system, it will be used if user defined function is not provided.

        Parameters
        ----------
        r : List
            The list contains all geometry variable of system.

        system : System
            The input system.

        """
        # TODO The method used to get shape from system should be generalized
        # in CosApp level instead of interface level.

        for child in system.children.values():
            SysExplorer.default_add_shape(r, child)

        def filter_shapes(p: "Port") -> bool:
            try:
                shape = p.shape
                visible = p.visible
            except AttributeError:
                return False
            else:
                return visible and shape is not None

        for port in filter(filter_shapes, system.outputs.values()):
            # print(f"{port.owner.full_name()}.{port.name}")
            if isinstance(port.shape, List):
                r.extend(port.shape)
            else:
                r.append(port.shape)

    def get_geometry(
        self, sys: System, get_all=False
    ) -> Union[
        OccParser, Dict[int, OccParser]
    ]:  # Dict[int, List[Dict[str, List[Union[int, float]]]]]:
        """Convert the open cascade object inside system into
        serializable data in order to send to front end.

        Parameters
        ----------
        sys : System
            The input system.

        Returns
        -------
        Dict[int, List[Dict[str, List[Union[int, float]]]]]
            A dictionary contains geometry data.
        """

        if self.geo_source is not None and "recorder" in self.geo_source:

            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                recorder = self._system().drivers[driver_name].recorder
            else:
                recorder = (
                    self._system()[".".join(path_list[0:-1])]
                    .drivers[driver_name]
                    .recorder
                )
            try:
                recorder_data = recorder.export_data()
            except:
                recorder_data = recorder.data
            n_index = len(recorder_data.index)
            if get_all:
                ret = OrderedDict()
                for idx in range(n_index):
                    r = []
                    for var_name in self.geo_source.get("variables", []):
                        column = recorder_data.get(var_name)
                        r.append(column[idx])
                    shape_data = OccParser(r)
                    ret[idx] = shape_data
                return ret
            else:
                r = []
                for var_name in self.geo_source.get("variables", []):
                    column = recorder_data.get(var_name)
                    r.append(column[n_index - 1])
                shape_data = OccParser(r)
                return shape_data

        else:

            add_shape = self.add_shape or SysExplorer.default_add_shape
            r = []

            try:
                add_shape(r, sys)
            except:
                pass

            return OccParser(r)

    def save_data(self, **kwarg) -> None:
        """Callback function used to get geometry data and log value
        after each time step.

        If a time driver exists in system, this function will be connected
        to the signal of corresponding recorder. Otherwise, il will be
        called after the main driver finished.

        Each time called, this function will read the geometry in system
        at current time step and store in `self.local_buffer`. The computation
        log of CosApp is used to update the `self.notification_msg`, which is
        observed in front-end.

        """
        time_ref = kwarg.get("time_ref")

        data = self.get_geometry(self._system())

        self.local_buffer[self.time_step] = data.threejs_data
        if self.old_log == "":
            log_value = self.log_stream.getvalue()
        else:
            log_value = "".join(self.log_stream.getvalue().rsplit(self.old_log))
        self.old_log = self.log_stream.getvalue()

        self.notification_msg = {
            "update": self.notification_msg["update"] + 1,
            "msg": f"Computed step {time_ref}",
            "log": log_value,
        }

        self.time_step += 1

    def __del__(self):
        if self.cosapp_thread is not None and self.cosapp_thread.is_alive():
            self.cosapp_thread.join()

        if self.cosapp_thread is not None:
            self.cosapp_thread.kill()

        if self.server_process is not None:
            self.stop_server()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.

from typing import Union, List
from traitlets import Unicode

from cosapp.systems import System

from cosapp_lab.widgets.base import BaseWidget
from cosapp_lab.widgets.chartwidget import ChartElement
from cosapp_lab.widgets.controllerwidget import ControllerComponent
from cosapp_lab.widgets.geometrywidget import GeometryComponent
from cosapp_lab.widgets.structurewidget import StructureComponent
from cosapp_lab.widgets.infowidget import SystemInfoComponent
from .component import WidgetView, DocumentView, SysExplorerComponent


class SysExplorer(BaseWidget):

    _model_name = Unicode("SysExplorerModel").tag(sync=True)
    _view_name = Unicode("SysExplorerView").tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):
        self.title = "SysExplorer widget"
        super().__init__(data, **kwargs)

    def init_component(self, **kwargs):
        self.register(SysExplorerComponent, **kwargs)
        self.register(ChartElement, **kwargs)
        self.register(StructureComponent, **kwargs)
        self.register(ControllerComponent, **kwargs)
        geom_options = dict(
            get_shapes = kwargs.pop("get_shapes", None),
            add_shape = kwargs.pop("add_shape", None),  # to be deprecated
            source = kwargs.pop("source", None),
        )
        self.register(GeometryComponent, **geom_options, **kwargs)
        self.register(WidgetView, **kwargs)
        self.register(DocumentView, **kwargs)
        self.register(SystemInfoComponent, **kwargs)



import os
from pathlib import Path
from distutils.dir_util import copy_tree
from pathlib import Path
from shutil import rmtree


def copy_jupyter_extension(root: str = None) -> str:
    if root is None:
        root = os.environ["CONDA_PREFIX"]
    ext_path = Path(root) / "share/jupyter/labextensions/cosapp_lab"
    if not ext_path.exists():
        ext_path.mkdir(exist_ok=True)
    else:
        for path in ext_path.glob("**/*"):
            if path.is_file():
                path.unlink()
            elif path.is_dir():
                rmtree(path)
    src_path = Path(__file__).parent.parent / "lab_static"
    copy_tree(str(src_path), str(ext_path))


if __name__ == "__main__":
    copy_jupyter_extension()



import importlib
import json
import logging
import sys, os
from pathlib import Path
from typing import Dict, List, Union, Optional
from cosapp_lab._version import __version__
from click.utils import echo
from cosapp_lab.module import CosappModuleServer
from cosapp_lab.utils.helpers import Suppressor, get_abs_dir, get_readme_from_path


from ..module.module_server import MODULE_MODE

CODE_TEMPLATE = """
from {{name}} import _cosapp_lab_load_module
args_list = {{args}}
_cosapp_lab_load_module(*args_list)
"""

COSAPP_CONFIG_DIR = Path.home() / ".cosapp.d"
OKGREEN = "\033[92m"
ENDC = "\033[0m"


def _get_module_json_data() -> Dict:
    """Helper function to read module data from CoSApp configuration folder"""
    config_path = COSAPP_CONFIG_DIR / "app"
    if not config_path.exists():
        config_path.mkdir(parents=True, exist_ok=True)
    json_path = config_path / "cosapp_module.json"
    if json_path.is_file():
        with open(json_path, "r") as f:
            old_data = json.load(f)
    else:
        old_data = {}

    return old_data


def cosapp_module_list() -> None:
    """Function to show all registered CosApp modules."""
    old_data = _get_module_json_data()

    if len(old_data) == 0:
        echo("No CoSApp module registered")
    else:
        echo(f"CoSApp Lab version {__version__}")
        for key, value in old_data.items():
            try:
                echo(
                    f'({value["meta"]["env_name"]}) {OKGREEN}{key}{ENDC} {value["meta"]["version"]} - {value["title"]}'
                )
            except KeyError:
                echo(f'{OKGREEN}{key}{ENDC}              {value["title"]}')


def cosapp_module_register(name_list: List[str]) -> Union[None, int]:
    """
    Function to register CoSApp modules to library.

    Parameters
    -----------
    - name_list :  List[str]
        List of module name to be registered.

    Returns
    --------
    Union[None, int]
        Returns the status of operation, `None` if failed, `1` if succeeded.

    """
    try:
        from nb_conda_kernels import CondaKernelSpecManager
    except ImportError:
        env_name = None
        conda_kernel_name = None
    else:
        logger = logging.getLogger()
        logger.setLevel(logging.ERROR)
        kernel_name = env_name = os.environ.get("CONDA_DEFAULT_ENV", None)
        conda_kernel_name = None
        if env_name == "base":
            kernel_name = "root"

        cksm = CondaKernelSpecManager(conda_only=True)
        specs = cksm.get_all_specs()
        for name in specs:
            conda_env_name = specs[name]["spec"]["metadata"]["conda_env_name"]
            if conda_env_name == kernel_name:
                conda_kernel_name = name
                break

    old_data = _get_module_json_data()
    status = None
    for name in name_list:
        try:
            logging.disable(logging.CRITICAL)
            module = importlib.import_module(f"{name}")
            init_function = getattr(module, "_cosapp_lab_load_module")
        except ModuleNotFoundError:
            echo(f"Can't import module {name}, skipping.")
            continue
        except AttributeError:
            echo(f"{name} is not a CoSApp module, skipping.")
            continue
        try:
            with Suppressor():
                init_function()
        except NotImplementedError:
            echo(f"Interface configuration is not implemented in {name}, skipping.")
            continue

        try:
            version = f"v{module.__version__}"
        except AttributeError:
            version = "Unknown"

        if name in old_data:
            echo(f"Module {name} already registered, skipping.")
            continue
        else:
            meta_function = getattr(module, "_cosapp_lab_module_meta")
            meta = meta_function()
            if meta is None:
                meta = {}
            if "version" not in meta:
                meta["version"] = version
            title = meta.get("title", name)
            code = CODE_TEMPLATE.replace("{{name}}", name)
            path = Path(module.__file__).parent
            readme = get_readme_from_path(path.parent)
            if readme is None:
                readme = get_readme_from_path(path)
            meta["readme"] = readme
            meta["env_name"] = env_name
            old_data[name] = {
                "code": code,
                "title": title,
                "meta": meta,
                "kernel": conda_kernel_name,
            }
            echo(f"{name} is successfully registered as CoSApp standalone module")
            status = 1
    
    json_path = COSAPP_CONFIG_DIR / "app" / "cosapp_module.json"
    with open(json_path, "w") as f:
        json.dump(old_data, f)

    return status


def cosapp_module_remove(nameList: List[str]) -> None:
    """
    Function to remove CoSApp modules from library.

    Parameters
    -----------
    - name_list :  List[str]
        List of module name to be removed.
    """
    old_data = _get_module_json_data()
    for name in nameList:
        if name in old_data:
            del old_data[name]
            echo(f"{name} is successfully removed from standalone module list")
        else:
            echo(f"{name} can not be found, skipping.")
            continue
    json_path = COSAPP_CONFIG_DIR / "app" / "cosapp_module.json"
    with open(json_path, "w+") as f:
        json.dump(old_data, f)


def cosapp_load_file(file: str) -> None:
    """
    Function to start CoSApp module server from a python file.

    Parameters
    -----------
    - file :  str
        Path to python file to be loaded.
    """
    file_path = get_abs_dir(Path().absolute(), file)
    try:
        with open(file_path, "r") as f:
            file_content = f.read()
    except FileNotFoundError:
        echo(f"No such file {file_path}")
        return
    CosappModuleServer.startup_code = file_content
    CosappModuleServer.title = f"CoSApp application - {file}"
    CosappModuleServer.module_mode = MODULE_MODE.SINGLE
    if "--arguments" in sys.argv:
        sys.argv.remove("--arguments")
    CosappModuleServer.launch_instance()


def cosapp_load_module(module: str, args: Optional[str] = None) -> None:
    """
    Function to start CoSApp module server from a registered CoSApp module.

    Parameters
    -----------
    - module :  str
        Name of registered CoSApp module.
    - args : Optional[str]
        List of arguments expressed in form of string. For example, a list of
        3 arguments `foo1`, `foo2`, `foo3` need to be expressed as "foo1 foo2 foo3"
    """
    if args is not None:
        str_list = [f'"{str(x.strip())}"' for x in args.split(" ")]
        args_list = f'[{",".join(str_list)}]'
    else:
        args_list = "[]"
    old_data = _get_module_json_data()
    CosappModuleServer.all_module_data = old_data
    CosappModuleServer.module_name = module
    if module == "__all__":
        if "--all" in sys.argv:
            sys.argv.remove("--all")
        if "-a" in sys.argv:
            sys.argv.remove("-a")
        CosappModuleServer.startup_code = ""
        CosappModuleServer.title = ""
        CosappModuleServer.module_mode = MODULE_MODE.MULTIPLE
        CosappModuleServer.launch_instance()
    else:
        if module in old_data:
            title = old_data[module]["title"]
            if title is None:
                title = f"CoSApp application - {module}"
            CosappModuleServer.startup_code = old_data[module]["code"]
            CosappModuleServer.title = title
            CosappModuleServer.args_list = args_list
            CosappModuleServer.module_mode = MODULE_MODE.SINGLE
            if "--arguments" in sys.argv:
                sys.argv.remove("--arguments")
            if "-args" in sys.argv:
                sys.argv.remove("-args")
            CosappModuleServer.launch_instance()
        else:
            echo(f"Module {module} is not registered yet, start registering")
            check = cosapp_module_register(module)
            if check == 1:
                cosapp_load_module(module, args)



# Import the first class found in each file having the same name
# as the file (test is case insensitive).
#
# Copyright (C) 2018-2021, CoSApp Team - Safran Group
#
# DO NOT MODIFY this file. It is automatically handled by CoSApp.
import logging

logger = logging.getLogger(__name__)


def _import_all_modules():
    """Dynamically imports all systems in the package."""
    import inspect
    import os

    all_objects = []
    globals_, locals_ = globals(), locals()

    # dynamically import all the package modules
    modules = set()
    json_files = set()
    for filename in os.listdir(os.path.dirname(__file__)):
        # process all python files in directory that don't start with underscore
        # (which also keeps this module from importing itself)
        modulename, ext = os.path.splitext(filename)
        if filename[0] != "_":
            if ext == ".py":
                modules.add(modulename)
            elif ext == ".json":
                json_files.add(filename)

    old_length = len(modules) + 1
    errors = {}
    while len(modules) and old_length > len(modules):
        old_length = len(modules)
        for modulename in modules.copy():
            package_module = ".".join([__name__, modulename])
            try:
                module = __import__(package_module, globals_, locals_, [modulename])
            except ModuleNotFoundError as err:
                raise err
            except ImportError as err:
                errors[modulename] = repr(err)
                continue

            # Only the class with the same name as the file will be imported
            found_class = False
            for obj_name in filter(lambda name: name[0] != "_", module.__dict__):
                found_class = modulename.lower() == obj_name.lower()
                obj = module.__dict__[obj_name]
                if found_class and inspect.isclass(
                    obj
                ):  # Check that the object found is a class
                    globals_[obj_name] = module.__dict__[obj_name]
                    all_objects.append(obj_name)
                    break

            if not found_class:
                logger.warning(
                    "File {}.py does not contain a class named {}. The file will be ignored."
                    "".format(package_module, modulename)
                )

            modules.discard(modulename)  # Remove module from the available list

    if modules:
        logger.warning("Failed to import from {} modules {}.".format(__name__, modules))
        for modulename in modules:
            logger.debug("{}: {}".format(modulename, errors[modulename]))

    from cosapp.systems import System
    from jsonschema import ValidationError

    def systemFactory(name: str, filename: str) -> System:
        obj = System.load(filename)
        obj.name = name
        return obj

    for json_file in json_files:  # Fake class behavior for System JSON file
        try:
            tmp_system = System.load(json_file)
        except (TypeError, AttributeError, ValidationError):
            logger.warning(
                'JSON file "{}" does not defined a CoSApp System.'.format(json_file)
            )
        else:
            obj_name = tmp_system.name.capitalize()
            globals_[obj_name] = lambda name: systemFactory(name, json_file)
            all_objects.append(obj_name)

    return all_objects


__all__ = _import_all_modules()



import logging
import shutil
import tempfile
import re
import webbrowser
import tornado.ioloop
import tornado.web
from jupyter_server.base.handlers import log

try:
    from nb_conda_kernels import CondaKernelSpecManager as KernelSpecManager
except ImportError:
    from jupyter_client.kernelspec import KernelSpecManager

import logging

import jinja2
from jupyter_server.services.kernels.handlers import (
    KernelHandler,
    MainKernelHandler,
)

try:
    from jupyter_server.services.kernels.websocket import KernelWebsocketHandler
except ImportError:
    # For jupyter_server < 2
    from jupyter_server.services.kernels.handlers import ZMQChannelsHandler as KernelWebsocketHandler

from jupyter_server.services.kernels.kernelmanager import MappingKernelManager
from jupyter_server.services.kernelspecs.handlers import MainKernelSpecHandler
from traitlets import Dict as tDict
from traitlets import Integer, Bool, Unicode, default
from traitlets.config.application import Application

from .handlers import (
    DEFAULT_STATIC_ROOT,
    MODULE_MODE,
    ROOT,
    TEMPLATE_ROOT,
    Default404Handler,
    MainEntryHandler,
    MainModuleHandler,
    StartupCodeHandler,
)

ENTRY_POINT = {MODULE_MODE.SINGLE: "main.html", MODULE_MODE.MULTIPLE: "index.html"}

_kernel_id_regex = r"(?P<kernel_id>\w+-\w+-\w+-\w+-\w+)"


class CosappModuleServer(Application):
    name = "cosapp_module"
    description = Unicode(
        """ 
        This launches a stand-alone server for CoSApp application.
        """
    )
    option_description = Unicode(
        """
        """
    )

    startup_code = Unicode(
        """
        """
    )

    args_list = Unicode("[]")

    title = Unicode(
        """
        CoSApp application
        """
    )

    module_name = Unicode("__all__")

    file = Unicode("file", config=True)

    port = Integer(6789, config=True, help="Port of the CoSApp server. Default 6789.")
    static_root = Unicode(
        str(DEFAULT_STATIC_ROOT),
        config=True,
        help="Directory holding static assets (HTML, JS and CSS files).",
    )
    aliases = {
        "port": "CosappModuleServer.port",
        "p": "CosappModuleServer.port",
        "static": "CosappModuleServer.static_root",
        "f": "CosappModuleServer.file",
        "url_prefix": "CosappModuleServer.url_prefix",
        "open_browser": "CosappModuleServer.open_browser",
    }
    connection_dir_root = Unicode(
        config=True,
        help=(
            "Location of temporary connection files. Defaults "
            "to system `tempfile.gettempdir()` value."
        ),
    )
    connection_dir = Unicode()

    all_module_data = tDict({}, help="Content of CoSApp module json file")

    module_mode = Integer(
        0, help="Module mode, 0 for single module, 1 for multiple modules"
    )

    url_prefix = Unicode("", config=True, help=("Prefix to append to handlers URL."))

    open_browser = Bool(True, config=True, help=("True if local web browser is used, False otherwise"))

    @default("connection_dir_root")
    def _default_connection_dir(self):
        return tempfile.gettempdir()

    @default("log_level")
    def _default_log_level(self):
        return logging.INFO

    def start(self):
        connection_dir = tempfile.mkdtemp(
            prefix="cosapp_module_", dir=self.connection_dir_root
        )
        kernel_spec_manager = KernelSpecManager(parent=self)
        kernel_manager = MappingKernelManager(
            parent=self,
            kernel_spec_manager=kernel_spec_manager,
            connection_dir=connection_dir,
            cull_idle_timeout=600,
            cull_interval=600,
            cull_connected=True,
        )
        if len(self.title) == 0:
            self.title = "CoSApp Application"
        HANDLE_PREFIX = r"%s" % self.url_prefix
        handlers = [
            (
                HANDLE_PREFIX + r"/cosapp/code",
                StartupCodeHandler,
                {
                    "code": self.startup_code,
                    "title": self.title,
                    "args": self.args_list,
                    "all_module_data": self.all_module_data,
                },
            ),
            (
                HANDLE_PREFIX + r"/module/([^/]+)",
                MainModuleHandler,
                {
                    "all_module_data": self.all_module_data,
                    "module_mode": self.module_mode,
                    "url_prefix": HANDLE_PREFIX,
                },
            ),
            (
                HANDLE_PREFIX + r"/api/kernels",
                MainKernelHandler,
            ),
            (HANDLE_PREFIX + r"/api/kernels/%s" % _kernel_id_regex, KernelHandler),
            (
                HANDLE_PREFIX + r"/api/kernels/%s/channels" % _kernel_id_regex,
                KernelWebsocketHandler,
            ),
            (HANDLE_PREFIX + r"/api/kernelspecs", MainKernelSpecHandler),
            (
                HANDLE_PREFIX + r"/static/(.*)",
                tornado.web.StaticFileHandler,
                {"path": self.static_root},
            ),
            (
                HANDLE_PREFIX + r"/",
                MainEntryHandler,
                {
                    "all_module_data": self.all_module_data,
                    "module_name": self.module_name,
                    "module_mode": self.module_mode,
                },
            ),
        ]
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATE_ROOT))
        app = tornado.web.Application(
            handlers,
            kernel_manager=kernel_manager,
            kernel_spec_manager=kernel_spec_manager,
            autoreload=False,
            jinja2_env=env,
            compress_response=False,
            allow_remote_access=True,
            allow_origin_pat=re.compile(".*"),
            default_handler_class=Default404Handler,
        )
        app.listen(self.port)
        url = f"http://localhost:{self.port}{self.url_prefix}/"
        self.log.info(f"CoSApp server listening on port {self.port}")
        self.log.info(f"Address: {url}")
        try:
            if self.open_browser:
                webbrowser.open(url)
            logging.getLogger("tornado.access").disabled = True
            tornado.ioloop.IOLoop.current().start()

        finally:
            self.log.info("Remove connection files")
            shutil.rmtree(connection_dir)


launch_server = CosappModuleServer.launch_instance

if __name__ == "__main__":
    CosappModuleServer.startup_code = ""
    CosappModuleServer.title = ""
    CosappModuleServer.module_mode = MODULE_MODE.MULTIPLE
    CosappModuleServer.launch_instance()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from typing import Any, Callable, Dict, List
from weakref import ReferenceType
from cosapp_lab.widgets.widgetview import IpyWidgetRender
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser
from cosapp_lab.widgets.base.base_component import BaseComponent
import re


class WidgetView(BaseComponent):
    name = "WidgetViewer"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List):
        if content["action"] == f"{self.name}::executeCode":
            code: str = content["payload"]["code"]
            title = content["payload"]["title"]
            sys_var = f"sys_{title}".replace("-", "_")
            check_code = re.search("[a-zA-Z]", code)
            if check_code is None or len(title) == 0 or "generate_widget" not in code:
                return
            global_dict = {
                sys_var: self.system,
                "IpyWidgetRender": IpyWidgetRender,
                "title": title,
            }
            exec(code, global_dict)
            code_with_arg = (
                f"IpyWidgetRender(title=title, children=[ generate_widget({sys_var})])"
            )
            code_without_arg = (
                f"IpyWidgetRender(title=title, children=[ generate_widget()])"
            )
            try:
                exec(code_with_arg, global_dict)
            except TypeError:
                exec(code_without_arg, global_dict)
            except:
                raise



"""
Information about the frontend package of the widgets.

Until decided otherwise, the frontend version is synchronized with the backend.
"""
from ._version import __version__

module_name = "cosapp_lab"
module_version = __version__



from .cosapp_init import cosapp_init
from .cosapp_jupyterlab import cosapp_jupyterlab, LOGO
from .cosapp_module import (
    cosapp_load_file,
    cosapp_module_register,
    cosapp_module_list,
    cosapp_module_remove,
    cosapp_load_module,
)

cosapp_script = {
    "init": cosapp_init,
    "lab": cosapp_jupyterlab,
    "cosapp_load_file": cosapp_load_file,
    "cosapp_load_module": cosapp_load_module,
    "cosapp_module_register": cosapp_module_register,
    "cosapp_module_list": cosapp_module_list,
    "cosapp_module_remove": cosapp_module_remove,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


"""
Widget container as geometry viewer panel
"""
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Union
from weakref import ReferenceType

from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser, OccParser
from cosapp_lab.widgets.base.base_component import BaseComponent

import warnings


class GeometryComponent(BaseComponent):
    name = "GeometryView"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        get_shapes = None,
        add_shape = None,  # to be deprecated
        source = None,
        **kwargs
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

        # TODO: remove block when `add_shape` is obsoleted
        if add_shape is not None:
            if get_shapes is None:
                def get_shapes(system: System) -> List[Any]:
                    """Encapsulation of legacy function `add_shape`"""
                    shapes = []
                    add_shape(shapes, system)
                    return shapes
                warnings.warn(
                    "`add_shape` has been replaced by `get_shapes`, and will be suppressed in future versions",
                    FutureWarning,
                )
            else:
                warnings.warn(
                    "Deprecated argument `add_shape` superseded by `get_shapes`",
                    RuntimeWarning,
                )
        self.get_shapes = get_shapes
        self.geo_source = source
        self.time_step = 0
        self.__init_connection()

    def __init_connection(self):
        """Initialize the connection between fontend - backend and
        between input system with the callbacks.
        """
        if self.geo_source is not None and "recorder" in self.geo_source:
            self._static = False
            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                recorder = self.system.drivers[driver_name].recorder
            else:
                recorder = (
                    self.system[".".join(path_list[0:-1])]
                    .drivers[driver_name]
                    .recorder
                )
            recorder.state_recorded.connect(self.save_data)
        else:
            driver_list = self.sys_data.get_time_driver()
            if len(driver_list) > 0:
                self._static = False
                time_driver = driver_list[0]
                if time_driver.recorder is not None:
                    time_driver.recorder.state_recorded.connect(self.save_data)
                else:
                    notification_recorder = time_driver.add_recorder(
                        DataFrameRecorder(includes=["_"])
                    )
                    notification_recorder.state_recorded.connect(self.save_data)
            else:
                self._static = True

    def save_data(self, **kwarg) -> None:
        """Callback function used to get geometry data
        after each time step.
        """
        data: OccParser = self.get_geometry()
        payload = {
            "threejs_data": data.threejs_data,
            "binary_position": data.binary_position,
            "time_step": self.time_step,
        }
        self.send(
            {"type": "GeometryView::geo_data", "payload": payload},
            [b.tobytes() for b in data.binary_data],
        )
        self.time_step += 1

    def send_initial_geometry(self) -> None:

        def pack_and_send(data, step=0, remaining=0) -> None:
            payload = {
                "threejs_data": data.threejs_data,
                "binary_position": data.binary_position,
                "time_step": step,
                "remaining": remaining,
            }
            self.send(
                {"type": "GeometryView::geo_data", "payload": payload},
                [b.tobytes() for b in data.binary_data],
            )

        if self.geo_source is not None and "recorder" in self.geo_source:
            all_data = self.get_geometry(get_all=True)
            remaining = len(all_data)
            for idx in all_data:
                remaining -= 1
                pack_and_send(all_data[idx], idx, remaining)
        else:
            data = self.get_geometry()
            if len(data.threejs_data) > 0:
                pack_and_send(data)

    def get_geometry(self, get_all=False) -> Union[OccParser, Dict[int, OccParser]]: 
        """Convert the open cascade objects inside `system` into
        serializable data in order to send to front end.

        Parameters
        ----------
        get_all : bool
            If `True`, returns all geometric data from recorder (if any)
            into a dictionary of `OccParser` objects.
            If `False` (default), returns geometric data collected in system.

        Returns
        -------
        Union[OccParser, Dict[int, OccParser]]
            `OccParser` object or dictionary thereof, depending on `get_all`.
        """
        if self.geo_source is not None and "recorder" in self.geo_source:
            # Argument `system` disregarded, here
            path_list = self.geo_source["recorder"].split(".")
            driver_name = path_list[-1]
            if len(path_list) == 1:
                system = self.system
            else:
                system = self.system[".".join(path_list[:-1])]
            recorder = system.drivers[driver_name].recorder
            columns = self.geo_source.get("variables", [])
            data = recorder.export_data().loc[:, columns]
            n_rows = len(data)

            get_occ_data = lambda i: OccParser(data.iloc[i].tolist())

            if get_all:
                return OrderedDict(
                    (i, get_occ_data(i))
                    for i in range(n_rows)
                )
            else:
                return get_occ_data(n_rows - 1)

        else:
            get_shapes = self.get_shapes or GeometryComponent.default_get_shapes
            try:
                shapes = get_shapes(self.system)
            except:
                shapes = []
            return OccParser(shapes)

    @staticmethod
    def default_get_shapes(system: System) -> List:
        """Default function used to extract geometry data from `system`,
        if no user-defined function is provided.

        Parameters
        ----------
        - system [System]:
            The system of interest.

        Returns
        -------
        - List:
            List containing all geometry variables of system.
        """
        # TODO The method used to get shapes from system should be generalized
        # in CosApp level instead of interface level.

        def is_geom_port(port: "Port") -> bool:
            try:
                shape = port.shape
                visible = port.visible
            except AttributeError:
                return False
            else:
                return visible and shape is not None

        def recursive_search(s: System, shapes: List):
            for child in s.children.values():
                recursive_search(child, shapes)
            
            for port in filter(is_geom_port, s.outputs.values()):
                if isinstance(port.shape, (list, tuple)):
                    shapes.extend(port.shape)
                else:
                    shapes.append(port.shape)

        shapes = []
        recursive_search(system, shapes)
        return shapes

    def computed_notification(self) -> None:
        """Callback function used to update geometry data from buffer
        and to emit update signal to front end.
        """
        if self._static:
            data = self.get_geometry()
            if len(data.threejs_data) > 0:
                payload = {
                    "threejs_data": data.threejs_data,
                    "binary_position": data.binary_position,
                    "time_step": 0,
                    "remaining": 0,
                }
                self.send(
                    {"type": "GeometryView::geo_data", "payload": payload},
                    [b.tobytes() for b in data.binary_data],
                )
        else:
            self.send(
                {
                    "type": "GeometryView::update_signal",
                    "payload": {},
                }
            )
            self.time_step = 0

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List):
        if content["action"] == "GeometryView::requestInitialGeometry":
            self.send_initial_geometry()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


import copy
import logging
import re
from io import StringIO
from typing import Any, Callable, Dict, List
from weakref import ReferenceType

from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser
from cosapp_lab.widgets.base.base_component import BaseComponent


class ControllerComponent(BaseComponent):
    name = "Controller"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)
        self.__init_connection()

    def __init_connection(self) -> None:
        """Initialize the connection between fontend - backend and
        between input system with the callbacks.
        """
        self.old_log = ""
        self.log_stream = StringIO()
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(stream=self.log_stream)
        handler.setFormatter(logging.Formatter(fmt="%(levelname)s:%(message)s"))
        logger.addHandler(handler)
        driver_list = self.sys_data.get_time_driver()
        if len(driver_list) > 0:
            self._static = False
            time_driver = driver_list[0]
            if time_driver.recorder is not None:
                time_driver.recorder.state_recorded.connect(self.update_log)
            else:
                notification_recorder = time_driver.add_recorder(
                    DataFrameRecorder(includes=["_"])
                )
                notification_recorder.state_recorded.connect(self.update_log)

    def update_log(self, **kwarg) -> None:
        """Callback function used to send log value to front end
        after each time step.

        If a time driver exists in system, this function will be connected
        to the signal of corresponding recorder. Otherwise, il will be
        called after the main driver finished.

        Each time called, this function will read the log of system
        at current time step and send a message with type
        `Controller::notification_msg` to front-end

        """
        time_ref = kwarg.get("time_ref")

        if self.old_log == "":
            log_value = self.log_stream.getvalue()
        else:
            log_value = "".join(self.log_stream.getvalue().rsplit(self.old_log))
        self.old_log = self.log_stream.getvalue()

        notification_msg = {
            "update": 1,
            "msg": f"Computed step {time_ref}",
            "log": log_value,
        }
        self.send({"type": "Controller::notification_msg", "payload": notification_msg})

    def computed_notification(self):

        server_log = self.log_stream.getvalue()
        self.log_stream.truncate(0)
        self.log_stream.seek(0)
        self.send(
            {
                "type": "Controller::update_signal",
                "payload": {"server_log": server_log},
            }
        )

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List):
        if content["action"] == "Controller::runSignal":
            self.sys_data.reset_variable_value()
            for key, variable_value in content["payload"].items():
                idx_group = re.search(r"\[(.*?)\]", key)
                if idx_group is not None:
                    var_idx = int(idx_group.group(1))
                    var_key = key.replace(idx_group.group(0), "")
                    sys_path = ".".join(var_key.split(".")[:-2])
                    port = var_key.split(".")[-2]
                    variable_name = var_key.split(".")[-1]

                    selected_system = self.sys_data.get_system_from_name(sys_path)
                    var_value = copy.deepcopy(selected_system[port][variable_name])
                    var_value[var_idx] = variable_value
                    self.sys_data.set_variable_value(
                        sys_path, port, variable_name, var_value
                    )
                else:
                    sys_path = ".".join(key.split(".")[:-2])
                    port = key.split(".")[-2]
                    variable_name = key.split(".")[-1]

                    self.sys_data.set_variable_value(
                        sys_path, port, variable_name, variable_value
                    )

            self.sys_data.run_system()



import json, copy, os
from typing import Any, Dict


def is_jsonable(x: Any) -> bool:
    """Helper function to check if input is jsonable

    Parameters
    ----------
    x : Any

    """

    try:
        json.dumps(x)
    except (TypeError, OverflowError):
        return False
    else:
        return True


def replicate_dict_structure(x: Dict) -> Dict:
    """Helper function to replicate a dict and remove all
    non-jsonable value

    Parameters
    ----------
    x : Dict

    Return
    -------
    New dict

    """
    new_dict = {}
    for key, current_val in x.items():
        if is_jsonable(current_val):
            new_dict[key] = copy.deepcopy(current_val)
        elif isinstance(current_val, dict):
            new_dict[key] = replicate_dict_structure(current_val)
        else:
            new_dict[key] = None
    return new_dict


def get_nonexistant_path(fname_path):
    if not os.path.exists(fname_path):
        return fname_path
    filename, extension = os.path.splitext(fname_path)
    get_name = lambda i: f"{filename}-{i}{extension}"
    i = 1
    new_fname = get_name(i)
    while os.path.exists(new_fname):
        i += 1
        new_fname = get_name(i)
    return new_fname



from typing import Any, Dict

import numpy as np
import pandas as pd

try:
    import plotly.graph_objs as go
except:
    go = None
from ipywidgets import HTML, Dropdown, HBox, Layout, Output, ToggleButtons, VBox
from traitlets import observe

from cosapp.systems import System
from cosapp.recorders import DataFrameRecorder
from .data import DataFrameRecorderViewer


class Statistics(DataFrameRecorderViewer):
    """Notebook widget to visualize DataFrameRecorder in statistical sense.

    Parameters
    ----------
    rec : cosapp.recorder.DataFrameRecorder
        Recorder containing the data to view
    live_update : int
        Point frequency at which the visualization is updated (0 == refresh at the end only)
    """

    def __init__(self, rec: DataFrameRecorder, live_update: int = 100, **kwargs):
        if go is None:
            raise ImportError(" this widget requires plotly")
        super(Statistics, self).__init__(rec, **kwargs)

        self._counter = 0  # Records counter
        self.live_update = max(0, live_update)
        parameters = self._compute_histogram(self.variables[0])
        self.widgets = w = {
            "graphics": ToggleButtons(
                value="Histogram",
                options={"": "Histogram", " ": "Scatter"},
                icons=["bar-chart", "line-chart"],
                layout=Layout(flex="0 0 auto", width="auto"),
            ),
            "plot": go.FigureWidget(
                data=[go.Histogram(**grp) for grp in parameters["data"]],
                layout=go.Layout(**parameters["layout"]),
            ),
            "axes": {
                "x": Dropdown(
                    options=self.variables,
                    description="x",
                    layout=Layout(flex="1 1 auto"),
                ),
                "y": Dropdown(
                    options=self.variables,
                    description="y",
                    layout=Layout(flex="1 1 auto", display="none"),
                ),
            },
            "debug": Output(),
        }

        # Connect internal widgets
        w["graphics"].observe(self.graphics_handler, "value")
        w["axes"]["x"].observe(self.x_handler, "value")
        w["axes"]["y"].observe(self.y_handler, "value")

        if self.live_update == 0:
            rec.state_recorded.disconnect(self._soft_update)

        # Connect to driver
        rec._owner.computed.connect(self.update)

        self.layout = Layout(
            height="100%", margin="5px", padding="0 0 10px 0", overflow_y="hidden"
        )
        self.children = [
            HBox(
                (
                    # w["graphics"],
                    HTML(
                        "<style>.cosapp-sidecar .js-plotly-plot { height: 100%; }</style>"
                    ),
                ),
                layout=Layout(flex="0 0 auto", overflow_y="hidden"),
            ),
            w["plot"],
            HBox(
                (w["axes"]["x"], w["axes"]["y"]),
                layout=Layout(
                    justify_content="center",
                    width="auto",
                    min_height="30px",
                    overflow_y="hidden",
                    flex="1 1 auto",
                ),
            ),
            VBox([w["debug"]], layout=Layout(overflow_y="auto")),
        ]

        # Force plot update
        self.update()

    def _compute_histogram(self, field: str) -> Dict[str, Any]:

        # List of percentiles to show with their corresponding colors
        percentiles = ["15%", "50%", "85%"]
        colors = ["grey", "black", "grey"]

        # Initialize the figure
        #   Compute the statiscal information for the first field
        if self.recorder.data.index.size == 0 or field == "index":
            return dict(
                data=list(),
                layout=dict(
                    title="Probability - " + field,
                    xaxis={"title": field},
                    yaxis={"title": "percent"},
                ),
            )
        else:
            n_section = 0
            shapes = list()
            data = list()
            for section, subdata in self.recorder.data[field].groupby("Section"):
                if len(subdata) == 0:
                    continue

                stats = subdata.describe(
                    percentiles=[0.15, 0.85]
                )  # see pandas.DataFrame.describe

                #   Create the vertical line for the percentiles

                for color, p in zip(colors, percentiles):
                    shapes.append(
                        go.layout.Shape(
                            type="line",
                            x0=stats[p],
                            x1=stats[p],
                            y0=0,
                            y1=1,
                            name=p,
                            yref="paper",  # Trick to get infinite length line
                            line={"color": color},
                        )
                    )

                data.append(
                    dict(
                        name=section,
                        x=subdata if field != "index" else [],
                        histnorm="percent",
                        opacity=0.8,
                        # marker={"color": "#f4950f"},
                    )
                )
                n_section += 1

            layout = dict(
                title="Probability - " + field,
                xaxis={"title": field},
                yaxis={"title": "percent"},
                shapes=shapes,
                margin=go.layout.Margin(l=70, r=50, b=40, t=50, pad=1),
            )
            if n_section > 1:
                layout["barmode"] = "overlay"

            # Create the figure object
            return dict(data=data, layout=layout)

    def _compute_scatter(self, x_label, y_label) -> Dict[str, Any]:
        percentiles = ["15%", "50%", "85%"]

        if self.recorder.data.index.size == 0:
            x_stats = pd.Series(np.zeros(len(percentiles)), index=percentiles)
            y_stats = x_stats
        else:
            x_stats = (
                (
                    self.recorder.data[x_label]
                    .groupby("Section")
                    .describe(percentiles=[0.15, 0.85])
                )
                if x_label != "index"
                else pd.Series(np.zeros(len(percentiles)), index=percentiles)
            )
            y_stats = (
                (
                    self.recorder.data[y_label]
                    .groupby("Section")
                    .describe(percentiles=[0.15, 0.85])
                )
                if y_label != "index"
                else pd.Series(np.zeros(len(percentiles)), index=percentiles)
            )

        return dict(
            data=dict(
                x=x_stats["50%"],
                y=y_stats["50%"],
                mode="markers",
                marker={"size": 8},
                error_x=dict(
                    type="data",
                    symmetric=False,
                    array=x_stats["85%"],
                    arrayminus=x_stats["15%"],
                ),
                error_y=dict(
                    type="data",
                    symmetric=False,
                    array=y_stats["85%"],
                    arrayminus=y_stats["15%"],
                ),
            ),
            layout=dict(
                title="",
                hovermode="closest",
                margin=go.layout.Margin(l=70, r=50, b=40, t=50, pad=1),
                xaxis=dict(
                    title=x_label,
                    showgrid=True,
                    zeroline=True,
                    showline=True,
                    nticks=10,
                    hoverformat=".3f",
                ),
                yaxis=dict(
                    title=y_label,
                    showgrid=True,
                    zeroline=True,
                    showline=True,
                    nticks=5,
                    hoverformat=".3f",
                ),
            ),
        )

    @observe("variables")
    def on_variables(self, change):
        """Handle changes in variable list."""
        for axis in ("x", "y"):
            dropdown = self.widgets["axes"][axis]
            v = dropdown.value
            dropdown.options = change["new"]
            dropdown.value = v

    def x_handler(self, change: Dict[str, Any]) -> None:
        """Handle *x* combobox change."""
        new = change["new"]
        if new:
            self.update({"x": new})

    def y_handler(self, change: Dict[str, Any]) -> None:
        """Handle *y* combobox change."""
        new = change["new"]
        if new:
            self.update({"y": new})

    def graphics_handler(self, change: Dict[str, Any]) -> None:
        """Handle click event graphics type."""
        new = change["new"]
        if new:
            # Update display widget
            self.widgets["axes"]["y"].layout.display = (
                "none" if new == "Histogram" else "flex"
            )
            # Clear figure
            self.widgets["plot"].data = []
            self.widgets["plot"].layout.shapes = []
            self.update({"graphics": new})

    def update(
        self, change: Dict[str, Any] = {"new": True}, force: bool = True, **kwargs
    ) -> None:
        """Handler to update the figure."""
        if self.recorder.paused:
            return

        with self.widgets["debug"]:
            if not force:
                self._counter += 1
                if self.live_update == 0 or self._counter % self.live_update != 0:
                    return
            else:
                self._counter = -1
            super(Statistics, self).update(change, **kwargs)

            plot = self.widgets["plot"]
            graphics = self.widgets["graphics"].value
            if graphics == "Histogram":
                field = self.widgets["axes"]["x"].value
                new_plot = self._compute_histogram(field)

                # Update the figure
                if plot.data and len(plot.data) == len(new_plot["data"]):
                    plot.update(
                        data=[{"x": grp["x"]} for grp in new_plot["data"]],
                        layout={
                            "title": "Probability - " + field,
                            "xaxis": {"title": field},
                            "shapes": new_plot["layout"]["shapes"],
                        },
                    )
                else:
                    # Force removing existing series - in case the number of trace changes
                    plot.data = []
                    for grp in new_plot["data"]:
                        plot.add_histogram(**grp)
                    plot.update(layout=new_plot["layout"])

            elif graphics == "Scatter":
                x_data = self.widgets["axes"]["x"].value
                y_data = self.widgets["axes"]["y"].value
                new_plot = self._compute_scatter(x_data, y_data)

                if plot.data:
                    plot.update(
                        data=[
                            {
                                "x": new_plot["data"]["x"],
                                "error_x": {
                                    "array": new_plot["data"]["error_x"]["array"],
                                    "arrayminus": new_plot["data"]["error_x"][
                                        "arrayminus"
                                    ],
                                },
                                "y": new_plot["data"]["y"],
                                "error_y": {
                                    "array": new_plot["data"]["error_y"]["array"],
                                    "arrayminus": new_plot["data"]["error_y"][
                                        "arrayminus"
                                    ],
                                },
                            }
                        ],
                        layout={"xaxis": {"title": x_data}, "yaxis": {"title": y_data}},
                    )
                else:
                    plot.add_scatter(**new_plot["data"])
                    plot.update(layout=new_plot["layout"])



from math import log10


def default_step(n: float) -> float:
    """Compute the default step for float number.

    The default step cannot be smaller than 5e-5.

    Parameters
    ----------
    n : float
        Base number to evaluate the step
    """
    if abs(n) < 1e-8:  # Close to zero
        return 0.5
    else:
        return 0.5 * 10 ** (max(int(log10(abs(n))) - 1, -4))



from cosapp_lab.widgets.base.base_widget import BaseWidget
from typing import List, Union
from cosapp.systems import System
from traitlets import Unicode


class ConnectionWidget(BaseWidget):

    _model_name = Unicode("ConnectionWidgetModel").tag(sync=True)
    _view_name = Unicode("ConnectionWidgetView").tag(sync=True)

    def __init__(self, data: Union[System, List[System]] = None, **kwargs):
        self.title = "Connection widget"
        super().__init__(data, **kwargs)

    def init_component(self, **kwargs):
        pass



from .cosapp_server import setup_cosapp_server_handlers

from notebook.notebookapp import NotebookWebApplication
from .cosapp_server import COSAPP_CONFIG_DIR


def _jupyter_server_extension_paths():
    """
    Function to declare Jupyter Server Extension Paths.
    """
    return [{"module": "cosapp_lab.server"}]


def load_jupyter_server_extension(nbapp: NotebookWebApplication):
    """
    Function to load Jupyter Server Extension.

    Parameters
    ----------
    nbapp : NotebookWebApplication
        Handle to the Notebook webserver instance.
    """
    setup_cosapp_server_handlers(nbapp.web_app)



from .cosapp_parser import CosappParser
from .cosapp_json_parser import CosappJsonParser
from .cosapp_object_parser import CosappObjectParser
from .occ_parser import OccParser
from .utils import is_jsonable, replicate_dict_structure, get_nonexistant_path



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) CoSApp Team.


from typing import Any, Callable, Dict, List
from weakref import ReferenceType
from cosapp_lab.widgets.widgetview import IpyWidgetRender
from cosapp.systems import System
from cosapp_lab.widgets.utils import CosappObjectParser
from cosapp_lab.widgets.base.base_component import BaseComponent
import markdown
from ipywidgets import HTML


class DocumentView(BaseComponent):
    name = "DocumentViewer"

    def __init__(
        self,
        data: "ReferenceType[System]",
        sys_data: CosappObjectParser,
        send_func: Callable,
        **kwargs,
    ):
        super().__init__(data=data, sys_data=sys_data, send_func=send_func, **kwargs)

    def _handle_button_msg(self, model: Any, content: Dict, buffers: List)-> None:
        if content["action"] == f"{self.name}::generateDocument":
            src: str = content["payload"]["source"]
            title: str = content["payload"]["title"]
            if len(title) == 0:
                return
            IpyWidgetRender(title=title, children=[self.generate_html(src)])


    def generate_html(self, src: str) -> HTML:
        html = markdown.markdown(src)
        return HTML(html)
